use crate::prelude::*;
use crate::rail_segment;
use crate::rail_segment::RailSegment;
use crate::rail_connector::RailConnector;
use crate::graph_coord::GraphCoord;
use std::collections::HashMap;
//use std::vec::Vec;
//use std::string::String;
use std::error;
use std::fmt;

#[derive(Debug)]
pub enum Error {
    IndexError(usize),
    SegmentError(rail_segment::Error),
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::IndexError(index) => write!(f, "Nonexistent segment index: {index}."),
            Self::SegmentError(err) => write!(f, "Segment error: {err}"),
        }
    }
}

impl error::Error for Error {}

pub struct RailGraph {
    segments: HashMap<usize, Box<dyn RailSegment>>,
    // segment_names: HashMap<String, usize>,
    
    min_available_index: usize,
}

impl RailGraph {
    pub fn new() -> Self {
        Self {
            segments: HashMap::<usize, Box::<dyn RailSegment>>::new(),
            //segment_names: {},
            min_available_index: 0,
        }
    }
    
    // definition
    
    pub fn get_segment(&self, index: usize) -> Result<&Box<dyn RailSegment>, Error> {
        match &self.segments.get(&index) {
            Some(ret) => Ok(ret),
            _         => Err(Error::IndexError(index)),
        }
    }
    
    pub fn add_segment(&mut self, segment: Box<dyn RailSegment>) -> usize {
        let ret = self.min_available_index;
        self.segments.insert(ret, segment);
        self.min_available_index += 1;
        ret
    }
    
    pub fn remove_segment(&mut self, index: usize) -> Result<bool, Error> {
        if !self.validate_index(index) {
            return Err(Error::IndexError(index));
        }
        
        self.isolate_segment(index);
        self.segments.remove(&index);
        Ok(true)
    }
    
    pub fn connect_endpoints(&mut self, index1: usize, endpoint1: usize, index2: usize, endpoint2: usize) -> Result<bool, Error> {
        // validate input first
        if !self.validate_index(index1) {
            return Err(Error::IndexError(index1));
        } else if !self.validate_index(index2) {
            return Err(Error::IndexError(index2));
        } else if !self.validate_endpoint(index1, endpoint1) {
            return Err(Error::SegmentError(rail_segment::Error::EndpointError(endpoint1)));
        } else if !self.validate_endpoint(index2, endpoint2) {
            return Err(Error::SegmentError(rail_segment::Error::EndpointError(endpoint2)));
        }
        
        // Needs improvement: silently fails with invalid endpoints.
        let connector1 = RailConnector::new(index2, endpoint2);
        let connector2 = RailConnector::new(index1, endpoint1);
        // Before we continue, make sure that anything that these endpoints are already connected to are disconnected.
        let old_connection1 = self.get_connection_from_index(index1, endpoint1);
        let old_connection2 = self.get_connection_from_index(index2, endpoint2);
        if let Ok(old_connection) = old_connection1 {
            match self.segments.get_mut(&old_connection.segment_index) {
                Some(old_segment) => {
                    if let Err(e) = old_segment.remove_connection(old_connection.endpoint_index) {
                        return Err(Error::SegmentError(e));
                    }
                },
                None              => {
                    return Err(Error::IndexError(old_connection.segment_index));
                },
            }
        }
        if let Ok(old_connection) = old_connection2 {
            match self.segments.get_mut(&old_connection.segment_index) {
                Some(old_segment) => {
                    if let Err(e) = old_segment.remove_connection(old_connection.endpoint_index) {
                        return Err(Error::SegmentError(e));
                    }
                },
                None              => {
                    return Err(Error::IndexError(old_connection.segment_index));
                },
            }
        }
        // Connect them now
        let segment1 = self.segments.get_mut(&index1).unwrap();
        let _ = segment1.add_connection(endpoint1, connector1);
        let segment2 = self.segments.get_mut(&index2).unwrap();
        let _ = segment2.add_connection(endpoint2, connector2);
        Ok(true)
    }
    
    pub fn disconnect_endpoint(&mut self, index: usize, endpoint: usize) -> Result<bool, Error> {
        if !self.validate_index(index) {
            return Err(Error::IndexError(index));
        } else if !self.validate_endpoint(index, endpoint) {
            return Err(Error::SegmentError(rail_segment::Error::EndpointError(endpoint)));
        }
        
        match self.get_connection_from_index(index, endpoint) {
            Ok(old_connection) => {
                match self.segments.get_mut(&old_connection.segment_index) {
                    Some(old_segment) => {
                        if let Err(e) = old_segment.remove_connection(old_connection.endpoint_index) {
                            return Err(Error::SegmentError(e));
                        }
                    },
                    None              => {
                        return Err(Error::IndexError(old_connection.segment_index));
                    },
                }
            },
            Err(e) => {
                return Err(e);
            },
        }
        
        let segment = self.segments.get_mut(&index).unwrap();
        match segment.remove_connection(endpoint) {
            Ok(_)  => Ok(true),
            Err(e) => Err(Error::SegmentError(e)),
        }
    }
    
    pub fn append_segment(&mut self,
                          new_segment: Box<dyn RailSegment>,
                          new_segment_endpoint: usize,
                          onto_segment: usize,
                          ont_segment_endpoint: usize) -> Result<bool, Error> {
        let index = self.add_segment(new_segment);
        self.connect_endpoints(index, new_segment_endpoint, onto_segment, ont_segment_endpoint)
    }
    
    // querying and traversal
    
    /**
     * Get a transform as described by the given coordinate.
     */
    pub fn transform_of_coord(&self, coord: &GraphCoord) -> Result<Transform3D, Error> {
        match self.get_segment(coord.segment) {
            Ok(segment) => {
                let ret = match coord.forward {
                    true => segment.sample_with_rotation(coord.subsegment, coord.offset),
                    false => segment.sample_with_rotation_reverse(coord.subsegment, coord.offset),
                };
                match ret {
                    Ok(tf) => Ok(tf),
                    Err(e) => Err(Error::SegmentError(e)),
                }
            },
            Err(e) => Err(e),
        }
    }
    
    /**
     * Get a coordinate defined by the previous coordinate, moved by an offset.
     * The offset can be negative and is relative to @coord.forward.
     */
    pub fn traverse(&self, coord: &GraphCoord, offset: f32) -> Result<GraphCoord, Error> {
        match self.get_segment(coord.segment) {
            Ok(segment) => {
                if segment.has_subsegment(coord.subsegment) {
                    let new_offset = coord.offset + match coord.forward {
                        true => offset,
                        false => -offset,
                    };
                    Ok(self.traverse_p(segment, coord.subsegment, coord.subsegment, new_offset, coord.forward))
                } else {
                    Err(Error::SegmentError(rail_segment::Error::IndexError(coord.subsegment)))
                }
            },
            Err(e) => Err(e),
        }
    }
    
    // private
    
    fn validate_index(&self, index: usize) -> bool {
        self.segments.contains_key(&index)
    }
    
    fn validate_endpoint(&self, index: usize, endpoint: usize) -> bool {
        match &self.segments.get(&index) {
            Some(segment) => {
                match segment.check_connection(endpoint) {
                    Err(rail_segment::Error::EndpointError(_)) => false,
                    _                                          => true,
                }
            },
            None          => false,
        }
    }
    
    fn get_connection_from_segment(segment: &Box<dyn RailSegment>, endpoint: usize) -> Result<RailConnector, Error> {
        match segment.get_connection(endpoint) {
            Ok(connection) => Ok(connection),
            Err(e)         => Err(Error::SegmentError(e)),
        }
    }
    
    fn get_connection_from_index(&self, index: usize, endpoint: usize) -> Result<RailConnector, Error> {
        match &self.segments.get(&index) {
            Some(segment) => Self::get_connection_from_segment(segment, endpoint),
            None          => Err(Error::IndexError(index)),
        }
    }
    
    fn isolate_segment(&mut self, index: usize) {
        // Get the old connections
        let mut connections = Vec::<RailConnector>::new();
        let segment = &self.segments[&index];
        for connection in segment.connection_iter() {
            connections.push(*connection);
        }
        
        // disconnect neighbors
        for connection in connections {
            let old_segment = self.segments.get_mut(&connection.segment_index).unwrap();
            let _ = old_segment.remove_connection(connection.endpoint_index);
        }
        
        // reflect the same change in the original segment.
        let segment = self.segments.get_mut(&index).unwrap();
        segment.remove_all_connections();
    }
    
    // Difference from the public version: offset is relative to the segment.
    // Segment index and segment both being arguments seems pretty redundant,
    // but the index isn't contained by the segment itself.
    // Maybe that should change.
    fn traverse_p(&self, segment: &Box<dyn RailSegment>, segment_index: usize, subsegment_index: usize, offset: f32, forward: bool) -> GraphCoord {
        // Use unwraps when I can because at this point, the graph should be valid.
        let segment_length = segment.get_length(subsegment_index).unwrap();
        if offset > segment_length {
            // We're off the upper end of the segment, so go to its neighbor.
            match segment.endpoint_from_subsegment(subsegment_index, true) {
                Ok(endpoint) => {
                    match segment.borrow_connection(endpoint) {
                        Ok(next_connection) => {
                            let next_segment = self.get_segment(next_connection.segment_index).unwrap();
                            let next_subsegment = next_segment.subsegment_from_endpoint(next_connection.endpoint_index).unwrap();
                            let next_endpoint_upper = next_segment.is_endpoint_upper(next_connection.endpoint_index).unwrap();
                            let next_forward = forward != next_endpoint_upper;
                            let next_offset = match next_endpoint_upper {
                                true => {
                                    let next_length = next_segment.get_length(next_subsegment).unwrap();
                                    next_length + segment_length - offset
                                },
                                false => offset - segment_length,
                            };
                            // Recurse with the next segment.
                            self.traverse_p(next_segment, next_connection.segment_index, next_subsegment, next_offset, next_forward)
                        },
                        Err(_) => {
                            // No connected segment, derail.
                            GraphCoord::new(segment_index, subsegment_index, offset, forward, true)
                        },
                    }
                },
                Err(_) => {
                    // Subsegment doesn't meet an endpoint, derail.
                    GraphCoord::new(segment_index, subsegment_index, offset, forward, true)
                },
            }
        } else if offset < 0.0 {
            // We're off the lower end of the segment, so go to its neighbor.
            match segment.endpoint_from_subsegment(subsegment_index, false) {
                Ok(endpoint) => {
                    match segment.borrow_connection(endpoint) {
                        Ok(next_connection) => {
                            let next_segment = self.get_segment(next_connection.segment_index).unwrap();
                            let next_subsegment = next_segment.subsegment_from_endpoint(next_connection.endpoint_index).unwrap();
                            let next_endpoint_upper = next_segment.is_endpoint_upper(next_connection.endpoint_index).unwrap();
                            let next_forward = forward == next_endpoint_upper;
                            let next_offset = match next_endpoint_upper {
                                true => {
                                    let next_length = next_segment.get_length(next_subsegment).unwrap();
                                    next_length + offset
                                },
                                false => -offset,
                            };
                            // Recurse with the next segment.
                            self.traverse_p(next_segment, next_connection.segment_index, next_subsegment, next_offset, next_forward)
                        },
                        Err(_) => {
                            // No connected segment, derail.
                            GraphCoord::new(segment_index, subsegment_index, offset, forward, true)
                        },
                    }
                },
                Err(_) => {
                    // Subsegment doesn't meet an endpoint, derail.
                    GraphCoord::new(segment_index, subsegment_index, offset, forward, true)
                },
            }
        } else {
            GraphCoord::new(segment_index, subsegment_index, offset, forward, false)
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::samplers::tangent_sampler::*;
    use crate::samplers::circular_sampler::*;
    use crate::samplers::sampler::*;
    use crate::simple_segment::*;
    use std::f32::consts::PI;
    
    /*fn compare_vectors(vec1: Vector3, vec2: Vector3, epsilon: f32, message: &str) {
        assert!((vec1.x - vec2.x).abs() <= epsilon && (vec1.y - vec2.y).abs() <= epsilon,
                "{message} Expected {vec2}, received {vec1}.")
    }*/
    
    /*fn compare_floats(float1: f32, float2: f32, epsilon: f32, message: &str) {
        assert!((float1 - float2).abs() <= epsilon,
                "{message} Expected {float2}, received {float1}.")
    }*/
    
    fn create_test_graph() -> RailGraph {
        let alignment1 = TangentAlignmentSampler::new(Vector2::new(0.0, 0.0), Vector2::new(1.0, 0.0));
        let elevation1 = TangentElevationSampler::new(0.0, 0.0, 1.0);
        let test_segment1 = Box::new(SimpleSegment::new(alignment1, elevation1));
        let alignment2 = CircularAlignmentSampler::new(0.0, PI * 0.5, Vector2::new(1.0, 1.0), 1.0, true);
        let elevation2 = TangentElevationSampler::new(0.0, 0.0, alignment2.length());
        let test_segment2 = Box::new(SimpleSegment::new(alignment2, elevation2));
        let alignment3 = TangentAlignmentSampler::new(Vector2::new(2.0, 1.0), Vector2::new(2.0, 2.0));
        let elevation3 = TangentElevationSampler::new(0.0, 0.0, 1.0);
        let test_segment3 = Box::new(SimpleSegment::new(alignment3, elevation3));
        
        let mut test_graph = RailGraph::new();
        test_graph.add_segment(test_segment1);
        let _ = test_graph.append_segment(test_segment2, 0, 0, 1);
        let _ = test_graph.append_segment(test_segment3, 0, 1, 1);
        return test_graph;
    }
    
    #[test]
    fn test_get_segment() {
        let mut test_graph = create_test_graph();
        
        // test successes
        assert!(match test_graph.get_segment(0) { Ok(_)=>true, Err(_)=>false, },
                "Segment 0 could not be retrieved.");
        assert!(match test_graph.get_segment(1) { Ok(_)=>true, _=>false, },
                "Segment 1 could not be retrieved.");
        assert!(match test_graph.get_segment(2) { Ok(_)=>true, _=>false, },
                "Segment 2 could not be retrieved.");
                
        // test failures
        assert!(match test_graph.get_segment(3) { Err(Error::IndexError(3)) =>true, _=>false, },
                "Segment was retrieved: index out of bounds.");
        _ = test_graph.remove_segment(1);
        assert!(match test_graph.get_segment(1) { Err(Error::IndexError(1)) =>true, _=>false, },
                "Segment was retrieved: deleted.");
    }
    
    #[test]
    fn test_add_segment() {
        let alignment = TangentAlignmentSampler::new(Vector2::new(2.0, 2.0), Vector2::new(2.0, 3.0));
        let elevation = TangentElevationSampler::new(0.0, 0.0, 1.0);
        let test_segment = Box::new(SimpleSegment::new(alignment, elevation));
        
        let mut test_graph = create_test_graph();
        let index = test_graph.add_segment(test_segment);
        assert_eq!(index, 3, "Segment was added with incorrect index.");
        assert_eq!(test_graph.segments.len(), 4, "Graph has incorrect number of segments.");
    }
    
    #[test]
    fn test_remove_segment() {
        let mut test_graph = create_test_graph();
        
        // test failures
        assert!(match test_graph.remove_segment(3) { Err(Error::IndexError(3))=>true, _=>false, },
                "Non-existent segment was removed.");
        
        // test successes
        assert!(match test_graph.remove_segment(1) { Ok(true)=>true, _=>false, },
                "Segment was not removed.");
        assert_eq!(test_graph.segments.len(), 2, "Graph has incorrect number of segments.");
        assert_eq!(test_graph.min_available_index, 3, "Incorrect min_available_index.");
        let first_segment = test_graph.get_segment(0).unwrap();
        assert!(!first_segment.check_connection(1).unwrap(), "Connection 1 wasn't broken.");
        let second_segment = test_graph.get_segment(2).unwrap();
        assert!(!second_segment.check_connection(0).unwrap(), "Connection 2 wasn't broken.");
        assert!(match test_graph.get_segment(1) { Err(Error::IndexError(1))=>true, _=>false, },
                "Segment that didn't exist was retrieved.");
    }
    
    #[test]
    fn test_connect_endpoints() {
        let mut test_graph = create_test_graph();
        
        // test failures
        assert!(match test_graph.connect_endpoints(3, 0, 2, 1) { Err(Error::IndexError(3))=>true, _=> false, },
                "Connection(3, 0) to (2, 1) was made.");
        assert!(match test_graph.get_segment(2).unwrap().borrow_connection(1) { Err(rail_segment::Error::ConnectionError(1))=>true, _=>false, },
                "Invalid connection at (2, 1) #1.");
        assert!(match test_graph.connect_endpoints(0, 2, 2, 1) { Err(Error::SegmentError(rail_segment::Error::EndpointError(2)))=>true, _=> false, },
                "Connection(0, 2) to (2, 1) was made.");
        assert!(match test_graph.get_segment(2).unwrap().borrow_connection(1) { Err(rail_segment::Error::ConnectionError(1))=>true, _=>false, },
                "Invalid connection at (2, 1) #2.");
        assert!(match test_graph.connect_endpoints(0, 0, 3, 1) { Err(Error::IndexError(3))=>true, _=> false, },
                "Connection(0, 0) to (3, 1) was made.");
        assert!(match test_graph.get_segment(0).unwrap().borrow_connection(0) { Err(rail_segment::Error::ConnectionError(0))=>true, _=>false, },
                "Invalid connection at (0, 0) #1.");
        assert!(match test_graph.connect_endpoints(0, 0, 2, 2) { Err(Error::SegmentError(rail_segment::Error::EndpointError(2)))=>true, _=> false, },
                "Connection(0, 0) to (2, 2) was made.");
        assert!(match test_graph.get_segment(0).unwrap().borrow_connection(0) { Err(rail_segment::Error::ConnectionError(0))=>true, _=>false, },
                "Invalid connection at (0, 0) #2.");
        assert!(match test_graph.connect_endpoints(0, 1, 3, 1) { Err(Error::IndexError(3))=>true, _=> false, },
                "Connection(0, 1) to (3, 1) was made.");
        assert!(match test_graph.get_segment(0).unwrap().borrow_connection(1) { Ok(_)=>true, _=>false, },
                "Removed connection at (0, 1) #1.");
        assert!(match test_graph.connect_endpoints(0, 1, 2, 2) { Err(Error::SegmentError(rail_segment::Error::EndpointError(2)))=>true, _=> false, },
                "Connection(0, 1) to (2, 2) was made.");
        assert!(match test_graph.get_segment(0).unwrap().borrow_connection(1) { Ok(_)=>true, _=>false, },
                "Removed connection at (0, 1) #2.");
        
        // test successes
        assert!(match test_graph.connect_endpoints(0, 0, 2, 1) { Ok(true)=>true, _=> false, },
                "Connection(0, 0) to (2, 1) was not made.");
        let first_segment = test_graph.get_segment(0).unwrap();
        let first_connection = first_segment.borrow_connection(0).unwrap();
        assert_eq!(first_connection.segment_index, 2, "Connection(0, 0) had the incorrect index.");
        assert_eq!(first_connection.endpoint_index, 1, "Connection(0, 0) had the incorrect endpoint.");
        let second_segment = test_graph.get_segment(2).unwrap();
        let second_connection = second_segment.borrow_connection(1).unwrap();
        assert_eq!(second_connection.segment_index, 0, "Connection(2, 1) had the incorrect index.");
        assert_eq!(second_connection.endpoint_index, 0, "Connection(2, 1) had the incorrect endpoint.");
        assert!(match test_graph.connect_endpoints(0, 1, 2, 0) { Ok(true)=>true, _=> false, },
                "Connection(0, 1) to (2, 0) was not made.");
        let first_segment = test_graph.get_segment(0).unwrap();
        let first_connection = first_segment.borrow_connection(1).unwrap();
        assert_eq!(first_connection.segment_index, 2, "Connection(0, 1) had the incorrect index.");
        assert_eq!(first_connection.endpoint_index, 0, "Connection(0, 1) had the incorrect endpoint.");
        let second_segment = test_graph.get_segment(2).unwrap();
        let second_connection = second_segment.borrow_connection(0).unwrap();
        assert_eq!(second_connection.segment_index, 0, "Connection(2, 0) had the incorrect index.");
        assert_eq!(second_connection.endpoint_index, 1, "Connection(2, 0) had the incorrect endpoint.");
        let middle_segment = test_graph.get_segment(1).unwrap();
        let first_connection = middle_segment.borrow_connection(0);
        assert!(match first_connection { Result::Err(_)=>true, _=>false },
                "Connection(1, 0) still exists.");
        let second_connection = middle_segment.borrow_connection(1);
        assert!(match second_connection { Result::Err(_)=>true, _=>false },
                "Connection(1, 1) still exists.")
    }
    
    #[test]
    fn test_disconnect_endpoint() {
        let mut test_graph = create_test_graph();
        
        // test failures
        assert!(match test_graph.disconnect_endpoint(3, 1) { Err(Error::IndexError(3))=>true, _=>false },
                "Invalid segment accessed.");
        assert!(match test_graph.disconnect_endpoint(1, 2) { Err(Error::SegmentError(rail_segment::Error::EndpointError(2)))=>true, _=>false },
                "Invalid endpoint disconnected.");
        assert!(match test_graph.disconnect_endpoint(0, 0) { Err(Error::SegmentError(rail_segment::Error::ConnectionError(0)))=>true, _=>false },
                "Unconnected endpoint disconnected.");
        
        // test successes
        assert!(match test_graph.disconnect_endpoint(0, 1) { Ok(true)=>true, _=>false },
                "Endpoint failed to disconnect.");
        let first_segment = test_graph.get_segment(0).unwrap();
        let first_connection = first_segment.borrow_connection(1);
        assert!(match first_connection { Result::Err(_)=>true, _=>false },
                "Connection(0, 1) still exists.");
        let middle_segment = test_graph.get_segment(1).unwrap();
        let middle_connection = middle_segment.borrow_connection(0);
        assert!(match middle_connection { Result::Err(_)=>true, _=>false },
                "Connection(1, 0) still exists.");
    }
    
    #[test]
    fn test_append_segment() {
        let test_graph = create_test_graph();
        let first_segment = test_graph.get_segment(0).unwrap();
        let second_segment = test_graph.get_segment(1).unwrap();
        let third_segment = test_graph.get_segment(2).unwrap();
        
        let connection01 = first_segment.borrow_connection(1).unwrap();
        assert_eq!(connection01.segment_index, 1, "Connection(0, 1) had the incorrect index.");
        assert_eq!(connection01.endpoint_index, 0, "Connection(0, 1) had the incorrect endpoint.");
        let connection10 = second_segment.borrow_connection(0).unwrap();
        assert_eq!(connection10.segment_index, 0, "Connection(1, 0) had the incorrect index.");
        assert_eq!(connection10.endpoint_index, 1, "Connection(1, 0) had the incorrect endpoint.");
        let connection11 = second_segment.borrow_connection(1).unwrap();
        assert_eq!(connection11.segment_index, 2, "Connection(1, 1) had the incorrect index.");
        assert_eq!(connection11.endpoint_index, 0, "Connection(1, 1) had the incorrect endpoint.");
        let connection20 = third_segment.borrow_connection(0).unwrap();
        assert_eq!(connection20.segment_index, 1, "Connection(2, 0) had the incorrect index.");
        assert_eq!(connection20.endpoint_index, 1, "Connection(2, 0) had the incorrect endpoint.");
    }
    
    #[test]
    fn test_transform_of_coord() {
        let test_graph = create_test_graph();
        let first_segment = test_graph.get_segment(0).unwrap();
        
        let coord1 = GraphCoord::new(0, 0, 0.5, true, false);
        let truth1 = first_segment.sample_with_rotation(0, 0.5).unwrap();
        let tf1 = test_graph.transform_of_coord(&coord1).unwrap();
        assert_eq!(tf1.origin, truth1.origin, "transform_of_coord origin comparison 1 failed.");
        assert_eq!(tf1.basis, truth1.basis, "transform_of_coord basis comparison 1 failed.");
        
        let coord2 = GraphCoord::new(0, 0, 0.5, false, false);
        let truth2 = first_segment.sample_with_rotation_reverse(0, 0.5).unwrap();
        let tf2 = test_graph.transform_of_coord(&coord2).unwrap();
        assert_eq!(tf2.origin, truth2.origin, "transform_of_coord origin comparison 2 failed.");
        assert_eq!(tf2.basis, truth2.basis, "transform_of_coord basis comparison 2 failed.");
        
        let coord3 = GraphCoord::new(3, 0, 0.5, true, false);
        let tf3 = test_graph.transform_of_coord(&coord3);
        assert!(match tf3 { Result::Err(Error::IndexError(3))=>true, _=>false },
                "Retrieved the transform of a coordinate with an invalid segment.");
        
        let coord4 = GraphCoord::new(0, 1, 0.5, true, false);
        let tf4 = test_graph.transform_of_coord(&coord4);
        assert!(match tf4 { Result::Err(Error::SegmentError(rail_segment::Error::IndexError(1)))=>true, _=>false },
                "Retrieved the transform of a coordinate with an invalid subsegment.");
        
        let coord5 = GraphCoord::new(0, 0, 1.5, true, false);
        let tf5 = test_graph.transform_of_coord(&coord5);
        assert!(match tf5 { Result::Err(Error::SegmentError(rail_segment::Error::OffsetError(1.5)))=>true, _=>false },
                "Retrieved the transform of a coordinate with an invalid offset.");
    }
    
    #[test]
    fn test_traverse() {
        // make a simple graph with 4 straight segments
        // Segments 0, 1, and 3 are "forward" and 2 is "backward".
        let alignment1 = TangentAlignmentSampler::new(Vector2::new(0.0, 0.0), Vector2::new(1.0, 0.0));
        let elevation1 = TangentElevationSampler::new(0.0, 0.0, 1.0);
        let test_segment1 = Box::new(SimpleSegment::new(alignment1, elevation1));
        let alignment2 = TangentAlignmentSampler::new(Vector2::new(1.0, 0.0), Vector2::new(2.0, 0.0));
        let elevation2 = TangentElevationSampler::new(0.0, 0.0, 1.0);
        let test_segment2 = Box::new(SimpleSegment::new(alignment2, elevation2));
        let alignment3 = TangentAlignmentSampler::new(Vector2::new(3.0, 0.0), Vector2::new(2.0, 0.0));
        let elevation3 = TangentElevationSampler::new(0.0, 0.0, 1.0);
        let test_segment3 = Box::new(SimpleSegment::new(alignment3, elevation3));
        let alignment4 = TangentAlignmentSampler::new(Vector2::new(31.0, 0.0), Vector2::new(4.0, 0.0));
        let elevation4 = TangentElevationSampler::new(0.0, 0.0, 1.0);
        let test_segment4 = Box::new(SimpleSegment::new(alignment4, elevation4));
        
        let mut test_graph = RailGraph::new();
        test_graph.add_segment(test_segment1);
        let _ = test_graph.append_segment(test_segment2, 0, 0, 1);
        let _ = test_graph.append_segment(test_segment3, 1, 1, 1);
        let _ = test_graph.append_segment(test_segment4, 0, 2, 0);
        
        //let first_segment = test_graph.get_segment(0).unwrap();
        
        let coord1 = GraphCoord::new(0, 0, 0.25, true, false);
        let truth1 = GraphCoord::new(0, 0, 0.75, true, false);
        let ret1 = test_graph.traverse(&coord1, 0.5).unwrap();
        assert_eq!(ret1.segment, truth1.segment, "Incorrect segment in test 1.");
        assert_eq!(ret1.subsegment, truth1.subsegment, "Incorrect subsegment in test 1.");
        assert_eq!(ret1.offset, truth1.offset, "Incorrect offset in test 1.");
        assert_eq!(ret1.forward, truth1.forward, "Incorrect forward in test 1.");
        assert_eq!(ret1.derailed, truth1.derailed, "Incorrect derailed in test 1.");
        
        let coord2 = GraphCoord::new(0, 0, 0.75, false, false);
        let truth2 = GraphCoord::new(0, 0, 0.25, false, false);
        let ret2 = test_graph.traverse(&coord2, 0.5).unwrap();
        assert_eq!(ret2.segment, truth2.segment, "Incorrect segment in test 2.");
        assert_eq!(ret2.subsegment, truth2.subsegment, "Incorrect subsegment in test 2.");
        assert_eq!(ret2.offset, truth2.offset, "Incorrect offset in test 2.");
        assert_eq!(ret2.forward, truth2.forward, "Incorrect forward in test 2.");
        assert_eq!(ret2.derailed, truth2.derailed, "Incorrect derailed in test 2.");
        
        let coord3 = GraphCoord::new(0, 0, 0.25, true, false);
        let truth3 = GraphCoord::new(3, 0, 0.25, true, false);
        let ret3 = test_graph.traverse(&coord3, 3.0).unwrap();
        assert_eq!(ret3.segment, truth3.segment, "Incorrect segment in test 3.");
        assert_eq!(ret3.subsegment, truth3.subsegment, "Incorrect subsegment in test 3.");
        assert_eq!(ret3.offset, truth3.offset, "Incorrect offset in test 3.");
        assert_eq!(ret3.forward, truth3.forward, "Incorrect forward in test 3.");
        assert_eq!(ret3.derailed, truth3.derailed, "Incorrect derailed in test 3.");
        
        let coord4 = GraphCoord::new(3, 0, 0.25, false, false);
        let truth4 = GraphCoord::new(0, 0, 0.25, false, false);
        let ret4 = test_graph.traverse(&coord4, 3.0).unwrap();
        assert_eq!(ret4.segment, truth4.segment, "Incorrect segment in test 4.");
        assert_eq!(ret4.subsegment, truth4.subsegment, "Incorrect subsegment in test 4.");
        assert_eq!(ret4.offset, truth4.offset, "Incorrect offset in test 4.");
        assert_eq!(ret4.forward, truth4.forward, "Incorrect forward in test 4.");
        assert_eq!(ret4.derailed, truth4.derailed, "Incorrect derailed in test 4.");
        
        let coord5 = GraphCoord::new(1, 0, 0.25, false, false);
        let truth5 = GraphCoord::new(2, 0, 0.75, true, false);
        let ret5 = test_graph.traverse(&coord5, -1.0).unwrap();
        assert_eq!(ret5.segment, truth5.segment, "Incorrect segment in test 5.");
        assert_eq!(ret5.subsegment, truth5.subsegment, "Incorrect subsegment in test 5.");
        assert_eq!(ret5.offset, truth5.offset, "Incorrect offset in test 5.");
        assert_eq!(ret5.forward, truth5.forward, "Incorrect forward in test 5.");
        assert_eq!(ret5.derailed, truth5.derailed, "Incorrect derailed in test 5.");
        
        let coord5 = GraphCoord::new(2, 0, 0.25, false, false);
        let truth5 = GraphCoord::new(3, 0, 0.75, true, false);
        let ret5 = test_graph.traverse(&coord5, 1.0).unwrap();
        assert_eq!(ret5.segment, truth5.segment, "Incorrect segment in test 5.");
        assert_eq!(ret5.subsegment, truth5.subsegment, "Incorrect subsegment in test 5.");
        assert_eq!(ret5.offset, truth5.offset, "Incorrect offset in test 5.");
        assert_eq!(ret5.forward, truth5.forward, "Incorrect forward in test 5.");
        assert_eq!(ret5.derailed, truth5.derailed, "Incorrect derailed in test 5.");
        
        let coord6 = GraphCoord::new(3, 0, 0.25, false, false);
        let truth6 = GraphCoord::new(2, 0, 0.75, true, false);
        let ret6 = test_graph.traverse(&coord6, 1.0).unwrap();
        assert_eq!(ret6.segment, truth6.segment, "Incorrect segment in test 6.");
        assert_eq!(ret6.subsegment, truth6.subsegment, "Incorrect subsegment in test 6.");
        assert_eq!(ret6.offset, truth6.offset, "Incorrect offset in test 6.");
        assert_eq!(ret6.forward, truth6.forward, "Incorrect forward in test 6.");
        assert_eq!(ret6.derailed, truth6.derailed, "Incorrect derailed in test 6.");
        
        let coord7 = GraphCoord::new(2, 0, 0.25, false, false);
        let truth7 = GraphCoord::new(1, 0, 0.75, true, false);
        let ret7 = test_graph.traverse(&coord7, -1.0).unwrap();
        assert_eq!(ret7.segment, truth7.segment, "Incorrect segment in test 7.");
        assert_eq!(ret7.subsegment, truth7.subsegment, "Incorrect subsegment in test 7.");
        assert_eq!(ret7.offset, truth7.offset, "Incorrect offset in test 7.");
        assert_eq!(ret7.forward, truth7.forward, "Incorrect forward in test 7.");
        assert_eq!(ret7.derailed, truth7.derailed, "Incorrect derailed in test 7.");
        
        let coord8 = GraphCoord::new(0, 0, 0.25, true, false);
        let truth8 = GraphCoord::new(3, 0, 1.25, true, true);
        let ret8 = test_graph.traverse(&coord8, 4.0).unwrap();
        assert_eq!(ret8.segment, truth8.segment, "Incorrect segment in test 8.");
        assert_eq!(ret8.subsegment, truth8.subsegment, "Incorrect subsegment in test 8.");
        assert_eq!(ret8.offset, truth8.offset, "Incorrect offset in test 8.");
        assert_eq!(ret8.forward, truth8.forward, "Incorrect forward in test 8.");
        assert_eq!(ret8.derailed, truth8.derailed, "Incorrect derailed in test 8.");
    }
}
